/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *   - Nicolas Brodu (initial base code, some modifications in SVN)        *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#ifndef CLASSIFIER_HiddenLayerPerceptron
#define CLASSIFIER_HiddenLayerPerceptron

#include <vector>

// soon to be in C++ standard
#include <boost/shared_ptr.hpp>

#include "../ClassifierInterface.h"

#include "MultiLayerPerceptron.h"

struct HiddenLayerPerceptronClassifier : public ClassifierInterface {
    HiddenLayerPerceptronClassifier();
    ~HiddenLayerPerceptronClassifier();
    bool setVFoldCrossValidation(int v);
    int getNumCVStep();
    // Support is true: performed either by directly weighting the error function, or by resampling with probability depending on the weights.
    // See: Holger Schwenk and Yoshua Bengio "Boosting Neural Networks", Neural Computation 12(8):1869-1887 (August 2000)
    bool supportWeightedInstance();
    double finishTraining(int innercvstep);
    double predict(const itpp::vec& instance);
    bool parseOption(int option, std::string option_arg);
    std::string getName();

protected:
    int nhidden, nhiddenMaxCV, nhiddenStepCV, nbpsteps;
    double learningRate;
    int crossValidation;
    int resample;
    boost::shared_ptr<MultiLayerPerceptron> mlp;
};

struct HiddenLayerPerceptronFactory : public ClassifierFactory {
    ClassifierInputType getType();
    ClassifierInterface* instanciateClassifier();
    void printClassifierHelpText(std::ostream& os);
    std::string getClassifierOptions();
    int getMethodId();
};

#endif
