/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License as        *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/





/*


NOTE: ONGOING WORK !
      This does not currently work and is much too slow
      Avoiding memory transfers would be good...


*/









#include <stdio.h>

/*
forward step:
applyWeightsBiasTransfer(input, hidden1)
applyWeightsBiasTransfer(hidden1, hidden2)
...
applyWeightsBias(hiddenN, output)
*/
__global__ void applyWeightsBiasTransfer(
    int ndata,
    int nresults,
    float* data,       // input data or hidden neurons
    float* weights,    // weights. row-major array of nresults*ndata float, nresults rows, ndata columns
    float* biases,     // biases. array of nresults bias
    float* result)     // result to be stored in the hidden layer
{
    int r = blockIdx.x * blockDim.x + threadIdx.x;
    if (r<nresults) {
        float* wp = weights + r * ndata;
        float res = biases[r];
        for (int d=0; d<ndata; ++d) res += wp[d] * data[d];
        result[r] = tanhf(res);
    }
}
// only difference: no transfer function here
__global__ void applyWeightsBias(
    int ndata,
    int nresults,
    float* data,       // input data or hidden neurons
    float* weights,    // weights. row-major array of nresults*ndata float, nresults rows, ndata columns
    float* biases,     // biases. array of nresults bias
    float* result)     // result to be stored in the hidden layer
{
    int r = blockIdx.x * blockDim.x + threadIdx.x;
    if (r<nresults) {
        float* wp = weights + r * ndata;
        float res = biases[r];
        for (int d=0; d<ndata; ++d) res += wp[d] * data[d];
        result[r] = res;
    }
}

/*
Batch back-propagation: accumulates gradients over all training set.
Note: gradient arrays shall be init to 0 before the first call, and divided by the number of calls after the last call.
*/
// grad for biases, no transfer
__global__ void batchBackPropB(
    int nresults,
    float* outputs,            // outputs of the network
    float* expectedOutputs,    // the targets that shall be produced
    float* gbiases,
    float instanceWeight)      // result weight biases. array of nresults bias
{
    int r = blockIdx.x * blockDim.x + threadIdx.x;
    if (r<nresults) {
        gbiases[r] += (outputs[r] - expectedOutputs[r]) * instanceWeight;
    }
}

// grad for last layer weights, no transfer
__global__ void batchBackPropW(
    int ndata,
    int nresults,
    float* outputs,            // outputs of the network
    float* expectedOutputs,    // the targets that shall be produced
    float* previousLayerData,  // input data or hidden neurons, ndata values
    float* gweights,
    float instanceWeight)          // result weight gradient. row-major array of nresults*ndata float, nresults rows, ndata columns
{
    int r = blockIdx.x * blockDim.x + threadIdx.x;
    int d = blockIdx.y * blockDim.y + threadIdx.y;
    if (r<nresults && d<ndata) {
        gweights[r * ndata + d] += (outputs[r] - expectedOutputs[r]) * instanceWeight * previousLayerData[d];
    }
}

// grad for last layer hidden value, no transfer
// shall not be cumulated over train examples, used as temporary variable for one back prop
__global__ void backPropGradRes(
    int nhidden,
    int noutputs,
    float* weightsHO, // in
    float* gbiasO,    // in
    float* ghidden)  // out
{
    int h = blockIdx.x * blockDim.x + threadIdx.x;
    if (h<nhidden) {
        float tmp = 0;
        for (int o=0; o<noutputs; ++o) tmp += weightsHO[o * nhidden + h] * gbiasO[o];
        ghidden[h] = tmp;
    }
}

// grad for biases, no transfer
__global__ void batchBackPropTransferB(
    int nresults,
    float* gradRes,      // input result gradients. For last array without transfer, this is predicted[r] - target[r]
    float* previousResultValues,  // input: previously computed result values
    float* gbiases)     // result weight biases. array of nresults bias
{
    int r = blockIdx.x * blockDim.x + threadIdx.x;
    if (r<nresults) {
        gbiases[r] += gradRes[r] * (1.0f - previousResultValues[r] * previousResultValues[r]);
    }
}
// grad for last layer weights, no transfer
__global__ void batchBackPropTransferW(
    int ndata,
    int nhidden,
    float* ghidden,      // in
    float* inputData,    // in
    float* hiddenValues, // in
    float* gweightsIH)  // out
{
    int h = blockIdx.x * blockDim.x + threadIdx.x;
    int d = blockIdx.y * blockDim.y + threadIdx.y;
    if (h<nhidden && d<ndata) {
        gweightsIH[h * ndata + d] += ghidden[h] * inputData[d] * (1.0f - hiddenValues[h] * hiddenValues[h]);
    }
}


__global__ void learnW(
    int ndata,
    int nresults,
    float learningRate,
    float* weights,
    float* gweights)
{
    int r = blockIdx.x * blockDim.x + threadIdx.x;
    int d = blockIdx.y * blockDim.y + threadIdx.y;
    if (r<nresults && d<ndata) {
        weights[r * ndata + d] -= learningRate * gweights[r * ndata + d];
    }
}

__global__ void learnB(
    int nresults,
    float learningRate,
    float* bias,
    float* gbias)
{
    int r = blockIdx.x * blockDim.x + threadIdx.x;
    if (r<nresults) {
        bias[r] -= learningRate * gbias[r];
    }
}


/*
Now host launcher functions
*/
extern "C" {

struct MLP_on_GPU {
    // GPU pointers, not main mem
    float* inputData;
    float* hiddenValues;
    float* outputValues;
    float* expectedOutputs;
    float* weightsIH;
    float* biasH;
    float* weightsHO;
    float* biasO;

    // backpropagation gradients
    float* gweightsIH;
    float* gbiasH;
    float* gweightsHO;
    float* gbiasO;
    float* ghidden;

    // general info
    int ninputs, nhiddens, noutputs;

    // internal data
    int batchCount;

    MLP_on_GPU(int ninputs, int nhiddens, int noutputs) {
        this->inputData = 0;
        this->hiddenValues = 0;
        this->outputValues = 0;
        this->expectedOutputs = 0;
        this->weightsIH = 0;
        this->biasH = 0;
        this->weightsHO = 0;
        this->biasO = 0;
        this->gweightsIH = 0;
        this->gbiasH = 0;
        this->gweightsHO = 0;
        this->gbiasO = 0;
        this->ghidden = 0;
        this->ninputs = ninputs;
        this->nhiddens = nhiddens;
        this->noutputs = noutputs;
        batchCount = 0;
    }
};

void check_err(const char* tag) {
    cudaError_t err = cudaGetLastError();
    if (err!=cudaSuccess) {
        printf("%s: %s\n",tag,cudaGetErrorString(err));
        exit(1);
    }
}

// create MLP structures on GPU memory, returns a handler "opaque" pointer to caller
void* mlp_cuda_init(int ninputs, int nhiddens, int noutputs, float (*randomWeightGenerator)() ) {
    // user struct, pointers to 0 for later detection of mem fail
    MLP_on_GPU* mlp = new MLP_on_GPU(ninputs,nhiddens,noutputs);
    float* randomData = 0;
    float scale = 1.0;
    cudaError_t err;

    // allocate GPU arrays
    cudaMalloc((void**)&mlp->inputData, ninputs*sizeof(float));
    err = cudaGetLastError(); if (err!=cudaSuccess) goto FAIL_DEALLOC;
    cudaMalloc((void**)&mlp->hiddenValues, nhiddens*sizeof(float));
    err = cudaGetLastError(); if (err!=cudaSuccess) goto FAIL_DEALLOC;
    cudaMalloc((void**)&mlp->outputValues, noutputs*sizeof(float));
    err = cudaGetLastError(); if (err!=cudaSuccess) goto FAIL_DEALLOC;
    cudaMalloc((void**)&mlp->expectedOutputs, noutputs*sizeof(float));
    err = cudaGetLastError(); if (err!=cudaSuccess) goto FAIL_DEALLOC;
    cudaMalloc((void**)&mlp->weightsIH, ninputs*nhiddens*sizeof(float));
    err = cudaGetLastError(); if (err!=cudaSuccess) goto FAIL_DEALLOC;
    cudaMalloc((void**)&mlp->biasH, nhiddens*sizeof(float));
    err = cudaGetLastError(); if (err!=cudaSuccess) goto FAIL_DEALLOC;
    cudaMalloc((void**)&mlp->weightsHO, nhiddens*noutputs*sizeof(float));
    err = cudaGetLastError(); if (err!=cudaSuccess) goto FAIL_DEALLOC;
    cudaMalloc((void**)&mlp->biasO, noutputs*sizeof(float));
    err = cudaGetLastError(); if (err!=cudaSuccess) goto FAIL_DEALLOC;
    cudaMalloc((void**)&mlp->gweightsIH, ninputs*nhiddens*sizeof(float));
    err = cudaGetLastError(); if (err!=cudaSuccess) goto FAIL_DEALLOC;
    cudaMalloc((void**)&mlp->gbiasH, nhiddens*sizeof(float));
    err = cudaGetLastError(); if (err!=cudaSuccess) goto FAIL_DEALLOC;
    cudaMalloc((void**)&mlp->gweightsHO, nhiddens*noutputs*sizeof(float));
    err = cudaGetLastError(); if (err!=cudaSuccess) goto FAIL_DEALLOC;
    cudaMalloc((void**)&mlp->gbiasO, noutputs*sizeof(float));
    err = cudaGetLastError(); if (err!=cudaSuccess) goto FAIL_DEALLOC;
    cudaMalloc((void**)&mlp->ghidden, nhiddens*sizeof(float));
    err = cudaGetLastError(); if (err!=cudaSuccess) goto FAIL_DEALLOC;

    // random init of weight arrays
    scale = 1.0 / sqrt(ninputs+1.0);
    randomData = new float[ninputs*nhiddens];
    for (int i=0; i<ninputs*nhiddens; ++i) randomData[i] = randomWeightGenerator() * scale;
    cudaMemcpy(mlp->weightsIH, randomData, ninputs*nhiddens*sizeof(float), cudaMemcpyHostToDevice);
    err = cudaGetLastError(); if (err!=cudaSuccess) goto FAIL_DEALLOC;
    delete[] randomData;
    scale = 1.0 / sqrt(nhiddens+1.0);
    randomData = new float[nhiddens];
    for (int i=0; i<nhiddens; ++i) randomData[i] = randomWeightGenerator() * scale;
    cudaMemcpy(mlp->biasH, randomData, nhiddens*sizeof(float), cudaMemcpyHostToDevice);
    err = cudaGetLastError(); if (err!=cudaSuccess) goto FAIL_DEALLOC;
    delete[] randomData;
    randomData = new float[nhiddens*noutputs];
    for (int i=0; i<nhiddens*noutputs; ++i) randomData[i] = randomWeightGenerator() * scale;
    cudaMemcpy(mlp->weightsHO, randomData, nhiddens*noutputs*sizeof(float), cudaMemcpyHostToDevice);
    err = cudaGetLastError(); if (err!=cudaSuccess) goto FAIL_DEALLOC;
    delete[] randomData;
    scale = 1.0 / sqrt(noutputs+1.0);
    randomData = new float[noutputs];
    for (int i=0; i<noutputs; ++i) randomData[i] = randomWeightGenerator() * scale;
    cudaMemcpy(mlp->biasO, randomData, noutputs*sizeof(float), cudaMemcpyHostToDevice);
    err = cudaGetLastError(); if (err!=cudaSuccess) goto FAIL_DEALLOC;
    delete[] randomData;

    return mlp;

FAIL_DEALLOC:
    printf("cudaMalloc: %s\n",cudaGetErrorString(err));
    if (mlp->inputData) cudaFree(mlp->inputData);
    if (mlp->hiddenValues) cudaFree(mlp->hiddenValues);
    if (mlp->outputValues) cudaFree(mlp->outputValues);
    if (mlp->expectedOutputs) cudaFree(mlp->expectedOutputs);
    if (mlp->weightsIH) cudaFree(mlp->weightsIH);
    if (mlp->biasH) cudaFree(mlp->biasH);
    if (mlp->weightsHO) cudaFree(mlp->weightsHO);
    if (mlp->biasO) cudaFree(mlp->biasO);
    if (mlp->gweightsIH) cudaFree(mlp->gweightsIH);
    if (mlp->gbiasH) cudaFree(mlp->gbiasH);
    if (mlp->gweightsHO) cudaFree(mlp->gweightsHO);
    if (mlp->gbiasO) cudaFree(mlp->gbiasO);
    if (mlp->ghidden) cudaFree(mlp->ghidden);
    if (randomData) delete[] randomData;
    exit(1);
    return 0;
}

void mlp_cuda_free(void* handler) {
    MLP_on_GPU* mlp = (MLP_on_GPU*)handler;
    if (mlp->inputData) cudaFree(mlp->inputData);
    if (mlp->hiddenValues) cudaFree(mlp->hiddenValues);
    if (mlp->outputValues) cudaFree(mlp->outputValues);
    if (mlp->expectedOutputs) cudaFree(mlp->expectedOutputs);
    if (mlp->weightsIH) cudaFree(mlp->weightsIH);
    if (mlp->biasH) cudaFree(mlp->biasH);
    if (mlp->weightsHO) cudaFree(mlp->weightsHO);
    if (mlp->biasO) cudaFree(mlp->biasO);
    if (mlp->gweightsIH) cudaFree(mlp->gweightsIH);
    if (mlp->gbiasH) cudaFree(mlp->gbiasH);
    if (mlp->gweightsHO) cudaFree(mlp->gweightsHO);
    if (mlp->gbiasO) cudaFree(mlp->gbiasO);
    if (mlp->ghidden) cudaFree(mlp->ghidden);
    check_err("mlp_cuda_free");
    delete mlp;
}

// give a host mem array of inputs,
// get results in the host mem array of outputs
void mlp_cuda_forward(void* handler, float* inputs, float* outputs) {
    MLP_on_GPU* mlp = (MLP_on_GPU*)handler;

    // transfer data to GPU
    cudaMemcpy(mlp->inputData, inputs, mlp->ninputs*sizeof(float), cudaMemcpyHostToDevice);
    check_err("mlp_cuda_forward 1");

    // propagate first layer
    dim3 dimBlock1(256);
    dim3 dimGrid1((mlp->nhiddens + dimBlock1.x - 1) / dimBlock1.x);
    applyWeightsBiasTransfer<<< dimGrid1, dimBlock1 >>>(
        mlp->ninputs,
        mlp->nhiddens,
        mlp->inputData,
        mlp->weightsIH,
        mlp->biasH,
        mlp->hiddenValues
    );
    check_err("mlp_cuda_forward 2");

    // propagate second layer
    dim3 dimBlock2(256);
    dim3 dimGrid2((mlp->nhiddens + dimBlock2.x - 1) / dimBlock2.x);
    applyWeightsBias<<< dimGrid2, dimBlock2 >>>(
        mlp->nhiddens,
        mlp->noutputs,
        mlp->hiddenValues,
        mlp->weightsHO,
        mlp->biasO,
        mlp->outputValues
    );
    check_err("mlp_cuda_forward 3");

    // get result back if required
    if (outputs) cudaMemcpy(outputs, mlp->outputValues, mlp->noutputs*sizeof(float), cudaMemcpyDeviceToHost);
    check_err("mlp_cuda_forward 4");
}

void mlp_cuda_backpropagate_init(void* handler) {
    MLP_on_GPU* mlp = (MLP_on_GPU*)handler;
    cudaMemset(mlp->gweightsIH, 0, mlp->ninputs*mlp->nhiddens*sizeof(float));
    check_err("mlp_cuda_backpropagate_init 1");
    cudaMemset(mlp->gbiasH, 0, mlp->nhiddens*sizeof(float));
    check_err("mlp_cuda_backpropagate_init 2");
    cudaMemset(mlp->gweightsHO, 0, mlp->nhiddens*mlp->noutputs*sizeof(float));
    check_err("mlp_cuda_backpropagate_init 3");
    cudaMemset(mlp->gbiasO, 0, mlp->noutputs*sizeof(float));
    check_err("mlp_cuda_backpropagate_init 4");
    mlp->batchCount = 0;
}

void mlp_cuda_backpropagate(void* handler, float* inputData, float* expectedOutputs, float instanceWeight) {
    MLP_on_GPU* mlp = (MLP_on_GPU*)handler;

    // first propagate inputs, outputs stay in GPU mem
    mlp_cuda_forward(handler, inputData, 0);

    // transfer expected data to GPU (the input data was transfered by the forward operation)
    cudaMemcpy(mlp->expectedOutputs, expectedOutputs, mlp->noutputs*sizeof(float), cudaMemcpyHostToDevice);
    check_err("mlp_cuda_backpropagate 1");

    // now backpropagate
    {
    dim3 dimBlock(16,16);
    dim3 dimGrid((mlp->noutputs + dimBlock.x - 1) / dimBlock.x, (mlp->nhiddens + dimBlock.y - 1) / dimBlock.y);
    batchBackPropW<<< dimGrid, dimBlock >>>(mlp->nhiddens, mlp->noutputs, mlp->outputValues, mlp->expectedOutputs, mlp->hiddenValues, mlp->gweightsHO, instanceWeight);
    check_err("mlp_cuda_backpropagate 2");
    }{
    dim3 dimBlock(256);
    dim3 dimGrid((mlp->noutputs + dimBlock.x - 1) / dimBlock.x);
    batchBackPropB<<< dimGrid, dimBlock >>>(mlp->noutputs, mlp->outputValues, mlp->expectedOutputs, mlp->gbiasO, instanceWeight);
    check_err("mlp_cuda_backpropagate 3");
    }{
    dim3 dimBlock(256);
    dim3 dimGrid((mlp->nhiddens + dimBlock.x - 1) / dimBlock.x);
    backPropGradRes<<< dimGrid, dimBlock >>>(mlp->nhiddens, mlp->noutputs, mlp->weightsHO, mlp->gbiasO, mlp->ghidden);
    check_err("mlp_cuda_backpropagate 4");
    }{
    dim3 dimBlock(16,16);
    dim3 dimGrid((mlp->nhiddens + dimBlock.x - 1) / dimBlock.x, (mlp->ninputs + dimBlock.y - 1) / dimBlock.y);
    batchBackPropTransferW<<< dimGrid, dimBlock >>>(mlp->ninputs, mlp->nhiddens, mlp->ghidden, mlp->inputData, mlp->hiddenValues, mlp->gweightsIH);
    check_err("mlp_cuda_backpropagate 5");
    }{
    dim3 dimBlock(256);
    dim3 dimGrid((mlp->nhiddens + dimBlock.x - 1) / dimBlock.x);
    batchBackPropTransferB<<< dimGrid, dimBlock >>>(mlp->nhiddens, mlp->ghidden, mlp->hiddenValues, mlp->gbiasH);
    check_err("mlp_cuda_backpropagate 6");
    }

    mlp->batchCount++;
}

void mlp_cuda_learn(void* handler, float learningRate) {
    MLP_on_GPU* mlp = (MLP_on_GPU*)handler;
    // average the gradients, then apply learning rate => equivalent to changing the learning rate !
    learningRate /= mlp->batchCount;
    {
    dim3 dimBlock(16,16);
    dim3 dimGrid((mlp->noutputs + dimBlock.x - 1) / dimBlock.x, (mlp->nhiddens + dimBlock.y - 1) / dimBlock.y);
    learnW<<< dimGrid, dimBlock >>>(mlp->nhiddens, mlp->noutputs, learningRate, mlp->weightsHO, mlp->gweightsHO);
    check_err("mlp_cuda_learn 1");
    }{
    dim3 dimBlock(256);
    dim3 dimGrid((mlp->noutputs + dimBlock.x - 1) / dimBlock.x);
    learnB<<< dimGrid, dimBlock >>>(mlp->noutputs, learningRate, mlp->biasO, mlp->gbiasO);
    check_err("mlp_cuda_learn 2");
    }{
    dim3 dimBlock(16,16);
    dim3 dimGrid((mlp->nhiddens + dimBlock.x - 1) / dimBlock.x, (mlp->ninputs + dimBlock.y - 1) / dimBlock.y);
    learnW<<< dimGrid, dimBlock >>>(mlp->ninputs, mlp->nhiddens, learningRate, mlp->weightsIH, mlp->gweightsIH);
    check_err("mlp_cuda_learn 3");
    }{
    dim3 dimBlock(256);
    dim3 dimGrid((mlp->nhiddens + dimBlock.x - 1) / dimBlock.x);
    learnB<<< dimGrid, dimBlock >>>(mlp->nhiddens, learningRate, mlp->biasH, mlp->gbiasH);
    check_err("mlp_cuda_learn 4");
    }
}

}
