/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *   - Nicolas Brodu (initial base code, some modifications in SVN)        *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#ifndef CLASSIFIER_NearestNeighbours
#include <vector>

#include "../ClassifierInterface.h"

struct NearestNeighboursClassifier : public ClassifierInterface {
    typedef double (*DistanceFunction)(const itpp::vec& instance1, const itpp::vec& instance2);

    NearestNeighboursClassifier(DistanceFunction dfunc, std::string distanceName, bool needSortSamples = false);
    ~NearestNeighboursClassifier();
    bool setVFoldCrossValidation(int v);
    int getNumCVStep();
    bool supportWeightedInstance();
    void train(const itpp::vec& instance, double classification, double instanceWeight = 1.0);
    double finishTraining(int innercvstep);
    double predict(const itpp::vec& instance);
    bool parseOption(int option, std::string option_arg);
    std::string getName();

    DistanceFunction dist;
    std::string distName;
    static double EuclidianDistance(const itpp::vec& x, const itpp::vec&y);
    static double KolmogorovSmirnovDistance(const itpp::vec& x, const itpp::vec&y);

    // Kp1 is K+1 so we can use the next neighbor for normalizing the distances to [0..1]
    // Trick from "Weighted k-Nearest-Neighbor Techniques and Ordinal Classification", Klaus Hechenbichler and Klaus Schliep, 2004
    int Kp1, KStepCV, KMaxCV;
    int crossValidation;
    bool sortSamples;

protected:
    double predictInternal(const itpp::vec& instance, int kp1orcv, std::vector<int>& tsetidx);
};


struct NearestNeighboursFactory : public ClassifierFactory {
    enum NNType {
        Euclidian,
        KolmogorovSmirnov
    };
    NNType type;

    NearestNeighboursFactory(NNType t = Euclidian);
    ~NearestNeighboursFactory();
    ClassifierInputType getType();
    ClassifierInterface* instanciateClassifier();
    void printClassifierHelpText(std::ostream& os);
    std::string getClassifierOptions();
    int getMethodId();
};

#endif
