/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *   - Nicolas Brodu (initial base code, some modifications in SVN)        *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#ifndef CLASSIFIER_SupportVectorMachines
#include <vector>

#include "../ClassifierInterface.h"

#include "svm.h"

struct SupportVectorMachinesClassifier : public ClassifierInterface {
    SupportVectorMachinesClassifier(libsvm::KernelType type);
    ~SupportVectorMachinesClassifier();
    bool setVFoldCrossValidation(int v); // true
    int getNumCVStep();
    bool supportWeightedInstance(); // false
    // train functions from parent OK: store instances for later use
    // finishTraining does the job
    double finishTraining(int innercvstep);
    double predict(const itpp::vec& instance);
    bool parseOption(int option, std::string option_arg);
    std::string getName();

    libsvm::svm_parameter svm_param;
    libsvm::svm_model* model;
    libsvm::svm_problem* problem;
    libsvm::svm_node* featurenode;

    double svmCost;
    double svmGamma;
    double svmBias;
    bool svmBiasIsSet;
    int svmDegree;
    double svmCostMaxCV;
    double svmGammaMaxCV;
    double svmBiasMaxCV;
    int svmDegreeMaxCV;
    int svmCostStepCV;
    int svmGammaStepCV;
    int svmBiasStepCV;
    int svmDegreeStepCV;

    int crossValidation;
};


struct SupportVectorMachinesFactory : public ClassifierFactory {

    libsvm::KernelType type;

    SupportVectorMachinesFactory(libsvm::KernelType t = libsvm::LINEAR);
    ~SupportVectorMachinesFactory();
    ClassifierInputType getType();
    ClassifierInterface* instanciateClassifier();
    void printClassifierHelpText(std::ostream& os);
    std::string getClassifierOptions();
    int getMethodId();
};



#endif
