/***************************************************************************
 *   Authors:                                                              *
 *   - Nicolas Brodu                                                       *
 *                                                                         *
 *   Copyrights:                                                           *
 *   - INRIA                                                               *
 *   - Nicolas Brodu (initial base code, some modifications in SVN)        *
 *                                                                         *
 *   This file is free software; you can redistribute it and/or modify     *
 *   it under the terms of the GNU General Public License as               *
 *   published by the Free Software Foundation; either version 2 of the    *
 *   License, or (at your option) any later version.                       *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU Lesser General Public License for more details.                   *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 *                                                                         *
 ***************************************************************************/

#ifndef CLASSIFIER_HyperMOG
#define CLASSIFIER_HyperMOG

#include <itpp/itbase.h>

struct HyperMOG {

    // Set these parameters directly (no accessor), then call prepareForEstimation
    itpp::vec v;
    itpp::vec beta;
    itpp::Array<itpp::mat> S;
    itpp::vec lambda;
    itpp::Array<itpp::vec> m;

    // some constants for posterior probability estimates: computed by prepareForEstimation
    itpp::vec classConst;
    itpp::Array<itpp::mat> invS;

    int N, D;

    void init(int nclasses, int ndimensions);

    // update the class constants with the new parameters
    void prepareForEstimation();

    double estimatePosteriorLogLikelihood(int classNumber, const itpp::vec& x);

    // Computes p(Class_i | x) -- should then be normalized
    double estimateUnNormalizedPosteriorProbability(int classNumber, const itpp::vec& x);

};

#endif
